import pytest

import chemformula.config
from chemformula.chemformula import ChemFormula

# Tests for functionality

@pytest.fixture(autouse=True, scope="module")
def enable_hydrogen_isotopes():
    chemformula.config.AllowHydrogenIsotopes = True


@pytest.mark.parametrize(
    "testinput1, testinput2, expected",
    [
        (ChemFormula("H2O", 0), ChemFormula("H", 1), ChemFormula("H3O", 1)),
        (ChemFormula("Ba", 2), ChemFormula("SO4", -2), ChemFormula("BaSO4", 0)),
        (ChemFormula("C4H6", 0), ChemFormula("C2H4", 0), ChemFormula("C6H10", 0)),
    ],
)
def test_addition(testinput1, testinput2, expected):
    assert testinput1 + testinput2 == expected


@pytest.mark.parametrize(
    "testinput1, testinput2, charge_testinput2, expected_isotope, expected_radioactive",
    [
        ("H2O", "D", 1, True, False),
        ("U", "O2", 0, False, True),
    ],
)
def test_addition_properties(testinput1, testinput2, charge_testinput2, expected_isotope, expected_radioactive):
    adduct = ChemFormula(testinput1) + ChemFormula(testinput2, charge_testinput2)
    assert adduct.contains_isotopes == expected_isotope
    assert adduct.is_radioactive == expected_radioactive


@pytest.mark.parametrize(
    "testinput1, testinput2, expected",
    [
        (ChemFormula("H3O", 1), ChemFormula("H", 1), ChemFormula("H2O", 0)),
        (ChemFormula("CH2ClCH3", 0), ChemFormula("HCl", 0), ChemFormula("CH2CH2", 0)),
    ],
)
def test_subtraction(testinput1, testinput2, expected):
    assert testinput1 - testinput2 == expected


@pytest.mark.parametrize(
    "testinput1, testinput2, expected",
    [
        (ChemFormula("C2H4", 0), 2, ChemFormula("C4H8", 0)),
        (2, ChemFormula("C2H4", 0), ChemFormula("C4H8", 0)),
        (ChemFormula("Na", 1), 4, ChemFormula("Na4", 4)),
    ],
)
def test_multiplication(testinput1, testinput2, expected):
    assert testinput1 * testinput2 == expected


# Tests for error handling


def test_addition_failed():
    with pytest.raises(TypeError):
        ChemFormula("H2O") + "H+"


def test_subtraction_failed_type():
    with pytest.raises(TypeError):
        ChemFormula("H2O") - "H+"


def test_subtraction_failed_negative_frequency():
    with pytest.raises(ValueError):
        ChemFormula("H2O") - ChemFormula("H3")


def test_subtraction_failed_empty_formula():
    with pytest.raises(ValueError):
        ChemFormula("H2O") - ChemFormula("OH2")


def test_multiplication_failed():
    with pytest.raises(TypeError):
        ChemFormula("C2H4") * 0.5
    with pytest.raises(TypeError):
        0.5 * ChemFormula("C2H4")
    with pytest.raises(TypeError):
        ChemFormula("C2H4") * "2"
    with pytest.raises(TypeError):
        "2" * ChemFormula("C2H4")
    with pytest.raises(TypeError):
        ChemFormula("C2H4") * ChemFormula("H2")


def test_multiplication_failed_non_positive_integer():
    with pytest.raises(ValueError):
        ChemFormula("C2H4") * -2
    with pytest.raises(ValueError):
        ChemFormula("C2H4") * 0



import pytest

from chemformula.chemformula import ChemFormulaDict, ChemFormulaDictFloat


# Tests for ChemFormulaDict
def test_chemformulardict_valid_set_get():
    cf_dict = ChemFormulaDict()
    cf_dict["C"] = 6
    assert cf_dict["C"] == 6
    # missing key returns default int 0
    assert isinstance(cf_dict["N"], int)
    assert cf_dict["N"] == 0

def test_chemformulardict_invalid_symbol_raises():
    cf_dict = ChemFormulaDict()
    with pytest.raises(ValueError):
        cf_dict["Xx"] = 1  # unknown element symbol

@pytest.mark.parametrize("bad_value", [-1, 1.5, "2"])
def test_chemformulardict_invalid_frequency_types(bad_value):
    cf_dict = ChemFormulaDict()
    with pytest.raises(ValueError):
        cf_dict["C"] = bad_value

def test_chemformulardict_init_with_invalid_mapping_raises():
    with pytest.raises(ValueError):
        ChemFormulaDict({"Xx": 1})

def test_chemformulardict_init_with_invalid_dict():
    with pytest.raises(TypeError):
        ChemFormulaDict("C6H12O6")


# Tests for ChemFormulaDictFloat
def test_chemformulardictfloat_valid_set_get_and_default():
    f = ChemFormulaDictFloat()
    f["O"] = 0.5
    assert isinstance(f["O"], float)
    assert abs(f["O"] - 0.5) < 1e-12
    # missing key returns default float 0.0
    assert isinstance(f["He"], float)
    assert f["He"] == 0.0

def test_chemformulardictfloat_accept_int_and_convert_to_float():
    cf_dict = ChemFormulaDictFloat()
    cf_dict["H"] = 2
    assert isinstance(cf_dict["H"], float)
    assert cf_dict["H"] == 2.0

def test_chemformulardictfloat_invalid_symbol_raises():
    cf_dict = ChemFormulaDictFloat()
    with pytest.raises(ValueError):
        cf_dict["Xx"] = 1.0  # unknown element symbol

def test_chemformulardictfloat_invalid_value_type_raises():
    cf_dict = ChemFormulaDictFloat()
    with pytest.raises(ValueError):
        cf_dict["O"] = "0.5"  # string not allowed

def test_chemformulardictfloat_init_with_invalid_mapping_raises():
    with pytest.raises(ValueError):
        ChemFormulaDictFloat({"Xx": 1.0})

def test_chemformulardictfloat_init_with_invalid_dict():
    with pytest.raises(TypeError):
        ChemFormulaDictFloat("C6H12O6")



from chemformula import ChemFormula

conjugated_diene = ChemFormula("C4H6")
dienophile = ChemFormula("C2H4")

diels_alder_adduct = conjugated_diene + dienophile

print("\n--- Adding ChemFormula Objects ---")
print(f" Butadiene {conjugated_diene.unicode} and ethylene {dienophile.unicode}"
      f" undergo a Diels-Alder reaction to form {diels_alder_adduct.unicode}.")
print(f" Molecular weight: {diels_alder_adduct.formula_weight:.2f} g/mol.")

# OUTPUT:
#
# --- Adding ChemFormula Objects ---
#  Butadiene C₄H₆ and ethylene C₂H₄ undergo a Diels-Alder reaction to form C₆H₁₀.
#  Molecular weight: 82.15 g/mol.
#


dichloroethane = ChemFormula("ClH2CCH2Cl")
hydrogen_chloride = ChemFormula("HCl")
vinyl_chloride = dichloroethane - hydrogen_chloride

print("\n--- Subtracting ChemFormula Objects ---")
print(f" Vinyl chloride {vinyl_chloride.hill_formula.unicode} is synthesized from dichloroethane"
      f" {dichloroethane.hill_formula.unicode} by elimination of hydrogen chloride"
      f" {hydrogen_chloride.hill_formula.unicode}.")
print(f" Molecular weight: {vinyl_chloride.formula_weight:.2f} g/mol.")

# OUTPUT:
#
# --- Subtracting ChemFormula Objects ---
#  Vinyl chloride C₂H₃Cl is synthesized from dichloroethane C₂H₄Cl₂ by elimination of hydrogen chloride ClH.
#  Molecular weight: 62.50 g/mol.
#


borane = ChemFormula("BH3")
diborane = 2 * borane
print("\n--- Multiplying ChemFormula Objects ---")
print(f" Diborane {diborane.hill_formula.unicode} is formed by the dimerization of two borane"
      f" {borane.hill_formula.unicode} molecules.")
print(f" Molecular weight of diborane: {diborane.formula_weight:.2f} g/mol.")

# OUTPUT:
#
# --- Multiplying ChemFormula Objects ---
#  Diborane B₂H₆ is formed by the dimerization of two borane BH₃ molecules.
#  Molecular weight of diborane: 27.67 g/mol.
#


ATP = ChemFormula("C10H12N5O13P3", -4)
water = ChemFormula("H2O")
dihydrogen_phosphate = ChemFormula("H2PO4", -1)

AMP = ATP + 2 * water - 2 * dihydrogen_phosphate

print("\n--- Arithmetics with ChemFormula Objects ---")
print(f" ATP ({ATP.hill_formula.unicode}) hydrolyzes with two water molecules"
      f" to AMP ({AMP.hill_formula.unicode}) and two inorganic phosphates ({dihydrogen_phosphate.unicode})"
      f" releasing energy for cellular processes.\n")

# OUTPUT:
#
# --- Arithmetics with ChemFormula Objects ---
#  ATP (C₁₀H₁₂N₅O₁₃P₃⁴⁻) hydrolyzes to AMP (C₁₀H₁₂N₅O₇P²⁻) and two inorganic phosphates (H₂PO₄⁻) releasing energy for cellular processes.
#
